
/******************************************************************************
* MODULE     : interpolate.hpp
* DESCRIPTION: Asymptotic interpolation
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __INTERPOLATE_HPP
#define __INTERPOLATE_HPP
#include "complexify.hpp"
#include "dense_vector.hpp"
namespace mmx {
#define TMPL template<typename R>
#define Vector vector<R>

/******************************************************************************
* Subroutines
******************************************************************************/

TMPL Vector
exp (const Vector& v) {
  nat i, d= dim (v);
  Vector r (R (0), d);
  for (i=0; i<d; i++)
    r(i)= exp (v[i]);
  return r;
}

TMPL Vector
log (const Vector& v) {
  nat i, d= dim (v);
  Vector r (R (0), d);
  for (i=0; i<d; i++)
    r(i)= log (v[i]);
  return r;
}

TMPL bool
good_approximation (const Vector& v, const Vector& w,
		    int require, double eps)
{
  //cout << "--------------------\n";
  //cout << "Compare " << v << "\nwith " << w << "\n";
  int d= dim (v), i;
  for (i= d-require; i<d; i++)
    if (abs (v[i] - w[i]) > eps * abs (v[i]))
      return false;
  return true;
}

TMPL void
normalize (Vector& v, int& first, int require, double eps) {
  nat d= dim(v);
  int last= first+ dim(v), n;
  if ((int) d < require) {
    v= Vector ();
    first= last;
  }
  else for (n= last-1; n >= first; n--)
    if (radius (v[n-first]) >= Radius_type(R) (eps) * abs_D (v[n-first])) {
      v= range (v, n+1-first, d);
      first= n+1;
      return;
    }
}

TMPL void
show_changes (const Vector& v, const Vector& r, int first) {
  cout << "-------------------------------------------------\n";
  int d= dim (v) - dim (r);
  for (int i=d; i<(int)dim(v); i++)
    cout << i+first << "\t" << v[i] << "\n\t" << r[i-d] << "\n";
}

/******************************************************************************
* Interpolating using standard functions
******************************************************************************/

TMPL Vector
basic_interpolate (const Vector& v, int first, int require) {
  //cout << "Basic interpolation of " << v << "\n";
  // Require a reasonable amount of terms
  nat d= dim(v);
  require= max (require, 10);
  if ((int) (dim (v)) < require) return Vector ();

  // Check whether we are constant
  Vector r (v[d-1], d);
  if (good_approximation (v, r, require, 0.1)) {
    //show_changes (v, r, first);
    cout << "constant(" << v[d-1] << ")";    
    return r;
  }

  // Initialize with x identity function
  int last= first+ dim(v), n;
  Vector x (R(0), d);
  for (n= first; n<last; n++)
    x (n-first)= R(n);

  // Compare with iterated logarithms
  double e= 1.0;
  int iterates=0;
  while (true) {
    // Check whether we are sufficiently close
    Vector r= (v[d-1] / x[d-1]) * x;
    if (good_approximation (v, r, require, iterates==0? 0.333: 0.1)) {
      cout << "constant(" << v[d-1]/x[d-1] << ")";
      if (iterates == 0) cout << "*n";
      else if (iterates == 1) cout << "*log(n)";
      else if (iterates == 2) cout << "*log(log(n))";
      else cout << "*log(log(log(n)))";
      return r;
    }

    // Continue with next iterated logarithm
    int start= (int) (e + 1.5);
    if (start >= last) return Vector ();
    if (start > first) {
      Vector rr= basic_interpolate (range (v, start-first, d), start, require);
      if (dim (rr) == 0) break;
      else return rr;
    }
    e= exp (e);
    iterates++;
    Vector y (R(0), d);
    for (nat i=0; i<d; i++)
      y(i)= log (x(i));
    x= y;
  }

  // Make positive (for real coefficients only)
  if (v[d-1] < 0) {
    //cout << "Opposite\n";
    Vector w= -v;
    Vector r= basic_interpolate (w, first, require);
    //if (dim (r) == 0) cout << " *";
    if (dim (r) == 0) return Vector ();
    //cout << "Back to invert\n";
    //show_changes (w, r, 0);
    cout << " opposite";
    return -r;
  }
  else {
    for (int i=d-1; i>=0; i--)
      if (v[i] <= 0)
	return basic_interpolate (range (v, i+1, d), first+i+1, require);
  }

  // Try taking logarithms
  {
    //cout << "Logarithm\n";
    Vector w= log (v);
    Vector r= basic_interpolate (w, first, require);
    //if (dim (r) == 0) cout << " *";
    if (dim (r) == 0) return Vector ();
    //cout << "Back to invert\n";
    //show_changes (w, r, 0);
    cout << " exp";
    return exp (r);
  }

  return Vector ();
}

/******************************************************************************
* Main interpolation routine
******************************************************************************/

TMPL Vector
interpolate (const Vector& v2, int first, int require, int depth) {
  Vector v= v2;
  if (depth <= 0) return basic_interpolate (v, first, require);
  //cout << "Interpolating " << v << "\n";

  // Normalization
  require= max (require, 10);
  normalize (v, first, require, 0.1);
  nat d= dim(v);
  int last= first+ d;
  if ((int) (dim (v)) < require) return Vector ();
  //cout << "Interpolate: " << v2[d-3] << ", " << v2[d-2] << ", " << v2[d-1] << "\n";

  // Invert if necessary
  if (abs (v[d-1]) < 1) {
    //cout << "Invert\n";
    Vector w= Vector (R(1), d) / v;
    Vector r= interpolate (w, first, require, depth);
    int start= first + d - dim (r);
    normalize (r, start, require, 0.5);
    //if (dim (r) == 0) cout << " *";
    if (dim (r) == 0) return basic_interpolate (v, first, require);
    //cout << "Back to invert\n";
    //show_changes (w, r, 0);
    cout << " invert";
    return Vector (R(1), dim (r)) / r;
  }

  // Smooth case
  //if (abs (v[d-1] - v[d-2]) < sqrt (R (last))) {
  if (abs (v[d-1] - v[d-2]) < sqrt (R (last)) * R(last)) {
    //cout << "Sum\n";
    Vector w= range (v, 0, d-1) - range (v, 1, d);
    Vector r= interpolate (w, first, require-1, depth-1);
    //if (dim (r) == 0) cout << " *";
    if (dim (r) == 0) return basic_interpolate (v, first, require);
    //cout << "Back to sum\n";
    //show_changes (w, r, 0);
    cout << " sum(" << v[d-1] << ")";
    int dr= dim(r);
    Vector s (R(0), dr+1);
    s(dr)= v[d-1];
    for (int i= dr-1; i >= 0; i--)
      s(i)= s(i+1) + r(i);
    return s;
  }

  // Steep case
  if (abs (v[d-1] / v[d-2]) < sqrt (R (last)) * R(last)) {
    //cout << "Product\n";
    Vector w= range (v, 0, d-1) / range (v, 1, d);
    Vector r= interpolate (w, first, require-1, depth-1);
    //if (dim (r) == 0) cout << " *";
    if (dim (r) == 0) return basic_interpolate (v, first, require);
    //cout << "Back to product\n";
    //show_changes (w, r, 0);
    cout << " product(" << v[d-1] << ")";
    int dr= dim(r);
    Vector s (R(0), dr+1);
    s(dr)= v[d-1];
    for (int i= dr-1; i >= 0; i--)
      s(i)= s(i+1) * r(i);
    return s;
  }

  // Make positive (for real coefficients only)
  if (v[d-1] < 0) {
    //cout << "Opposite\n";
    Vector w= -v;
    Vector r= interpolate (w, first, require, depth);
    //if (dim (r) == 0) cout << " *";
    if (dim (r) == 0) return Vector ();
    //cout << "Back to invert\n";
    //show_changes (w, r, 0);
    cout << " opposite";
    return -r;
  }
  else {
    for (int i=d-1; i>=0; i--)
      if (v[i] <= 0)
	return interpolate (range (v, i+1, d), first+i+1, require, depth);
  }

  // Try taking logarithms
  {
    //cout << "Logarithm\n";
    Vector w= log (v);
    Vector r= interpolate (w, first, require, depth-1);
    //if (dim (r) == 0) cout << " *";
    if (dim (r) == 0) return Vector ();
    //cout << "Back to invert\n";
    //show_changes (w, r, 0);
    cout << " exp";
    return exp (r);
  }
}

#undef TMPL
#undef Vector
} // namespace mmx
#endif // __MMX_INTERPOLATE_HPP
