
/******************************************************************************
* MODULE     : routine.hpp
* DESCRIPTION: Abstract routines for generic expressions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __ROUTINE_HPP
#define __ROUTINE_HPP
#include "evaluator.hpp"
#include "generic_internal.hpp"
#include "array.hpp"
#include "exception.hpp"
namespace mmx {
class environment;

/******************************************************************************
* Abstraction
******************************************************************************/

template<typename T>
struct generic_converter {
  static inline T make_concrete (const generic& x) {
    return as<T,generic> (x); }
  static inline generic make_abstract (const T& x) {
    // NOTE: returning as<generic,T> does not seem to call the appropriate
    // converter in the case when T=bool. A compiler bug?
    return new generic_concrete_rep<T> (x); }
};

STMPL
struct generic_converter<generic> {
  static inline generic make_concrete (const generic& x) {
    return x; }
  static inline generic make_abstract (const generic& x) {
    return x; }
};

template<typename T> inline T
make_concrete (const generic& x) {
  return generic_converter<T>::make_concrete (x);
}

template<typename T> inline generic
make_abstract (const T& x) {
  return generic_converter<T>::make_abstract (x);
}

/******************************************************************************
* Abstract generic routines
******************************************************************************/

class routine;
class routine_rep: public rep_struct {
public:
  generic name;
public:
  inline routine_rep (const generic& name2): name (name2) {}
  virtual inline ~routine_rep () {}
  virtual generic apply () const;
  virtual generic apply (const generic& g1) const;
  virtual generic apply (const generic& g1, const generic& g2) const;
  virtual generic xapply (const generic& x) const;
  virtual generic apply (const array<generic>& args) const;
  virtual array<nat> signature () const;
  virtual void overload (const routine& fun) const;
};

class routine {
INDIRECT_PROTO (routine, routine_rep)
  inline routine (): rep (NULL) {}
  inline bool operator == (const routine& fun) const {
    return rep == fun.rep; }
  inline bool operator != (const routine& fun) const {
    return rep != fun.rep; }
  friend inline bool nil (const routine& fun);
  friend inline nat hash (const routine& fun);
};
INDIRECT_NULL_IMPL (routine, routine_rep)

inline bool nil (const routine& fun) {
  return fun.rep == NULL; }
inline nat hash (const routine& fun) {
  return ((nat) ((void*) fun.rep)) >> 3; }
inline generic flatten (const routine& fun) {
  if (nil (fun)) return "nil"; else return fun->name; }

PRINT_SUGAR(,routine);

/******************************************************************************
* Concrete generic routines
******************************************************************************/

template<typename D>
class nullary_routine_rep: public routine_rep {
  typedef D (*function_type) ();
  function_type fun;
public:
  nullary_routine_rep (const generic& name, function_type fun2):
    routine_rep (name), fun (fun2) {}
  generic apply () const {
    return make_abstract<D> (fun ());
  }
  generic xapply (const generic& g) const {
    assert (N(g)==1);
    return apply ();
  }
  array<nat> signature () const {
    return array<nat> (type_information<D>::id);
  }
};

template<typename D> routine
nullary_routine (const generic& name, D (*fun) ()) {
  return new nullary_routine_rep<D> (name, fun);
}

extern generic type_name (nat id);

template<typename D, typename S1>
class unary_routine_rep: public routine_rep {
  typedef D (*function_type) (const S1&);
  function_type fun;
public:
  unary_routine_rep (const generic& name, function_type fun2):
    routine_rep (name), fun (fun2) {}
  generic apply (const generic& x1) const {
    return make_abstract<D> (fun (make_concrete<S1> (x1)));
  }
  generic xapply (const generic& g) const {
    assert (N(g)==2);
    return apply (g[1]);
  }
  array<nat> signature () const {
    return array<nat> (type_information<D>::id, type_information<S1>::id);
  }
};

template<typename D, typename S1> routine
unary_routine (const generic& name, D (*fun) (const S1&)) {
  return new unary_routine_rep<D,S1> (name, fun);
}

template<typename D, typename S1, typename S2>
class binary_routine_rep: public routine_rep {
  typedef D (*function_type) (const S1&, const S2&);
  function_type fun;
public:
  binary_routine_rep (const generic& name, function_type fun2):
    routine_rep (name), fun (fun2) {}
  generic apply (const generic& x1, const generic& x2) const {
    return make_abstract<D> (fun (make_concrete<S1> (x1),
				  make_concrete<S2> (x2)));
  }
  generic xapply (const generic& g) const {
    assert (N(g)==3);
    return apply (g[1], g[2]);
  }
  array<nat> signature () const {
    return array<nat> (type_information<D>::id,
		       type_information<S1>::id,
		       type_information<S2>::id);
  }
};

template<typename D, typename S1, typename S2> routine
binary_routine (const generic& name, D (*f) (const S1&, const S2&)) {
  return new binary_routine_rep<D,S1,S2> (name, f);
}

template<typename D, typename S1, typename S2, typename S3>
class ternary_routine_rep: public routine_rep {
  typedef D (*function_type) (const S1&, const S2&, const S3&);
  function_type fun;
public:
  ternary_routine_rep (const generic& name, function_type fun2):
    routine_rep (name), fun (fun2) {}
  generic xapply (const generic& g) const {
    assert (N(g)==4);
    return make_abstract<D> (fun (make_concrete<S1> (g[1]),
				  make_concrete<S2> (g[2]),
				  make_concrete<S3> (g[3])));
  }
  array<nat> signature () const {
    return array<nat> (type_information<D>::id,
		       type_information<S1>::id,
		       type_information<S2>::id,
		       type_information<S3>::id);
  }
};

template<typename D, typename S1, typename S2, typename S3> routine
ternary_routine (const generic& name,
		 D (*f) (const S1&, const S2&, const S3&)) {
  return new ternary_routine_rep<D,S1,S2,S3> (name, f);
}

/******************************************************************************
* Constructing objects from generators
******************************************************************************/

template<typename T>
class from_iterator_routine_rep: public routine_rep {
public:
  from_iterator_routine_rep (const generic& name):
    routine_rep (name) {}
  generic apply (const generic& x) const {
    iterator<T> it= as<iterator<T> > (x);
    array<generic> t= array<generic> (name);
    while (busy (it)) {
      generic val= make_abstract<T> (*it); ++it;
      if (is<exception> (val)) return val;
      t << val;
    }
    return current_ev->xapply (array_to_compound (t));
  }
  generic xapply (const generic& g) const {
    assert (N(g) == 2);
    return apply (g[1]);
  }
  array<nat> signature () const {
    return array<nat> ((nat) 0, type_information<iterator<T> >::id);
  }
};

template<typename T> routine
from_iterator_routine (const generic& name) {
  return new from_iterator_routine_rep<T> (name);
}

/******************************************************************************
* Other types of routines
******************************************************************************/

routine default_routine (const generic& name);
routine identity_routine (const array<nat>& sig);
routine compose (const routine& fun, const array<routine>& args);
routine overloaded_routine (const generic& name, const environment& env);

} // namespace mmx
#endif // __ROUTINE_HPP
