
/******************************************************************************
* MODULE     : environment.cpp
* DESCRIPTION: Environments for the base evaluator
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "environment.hpp"
#include "literal.hpp"
#include "routine.hpp"
#include "alias.hpp"
namespace mmx {

/******************************************************************************
* Environment access
******************************************************************************/

generic
environment_rep::name () const {
  array<generic> vars (entries (bindings));
  if (nil (next)) return gen ("global", vars);
  else return gen ("local", append (vars, array<generic> (next->name ())));
}

bool
environment_rep::contains (const generic& var) const {
  if (bindings->contains (var)) return true;
  else if (nil (next)) return false;
  else if (next->contains (var)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    me->bindings (var)= next[var];
    return true;
  }
  else return false;
}

bool
environment_rep::read (const generic& var, generic& val) const {
  if (bindings->read (var, val)) return true;
  else if (nil (next)) return false;
  else if (next->read (var, val)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    me->bindings (var)= val;
    return true;
  }
  else return false;
}

generic
environment_rep::get (const generic& var) const {
  generic val;
  if (bindings->read (var, val)) return val;
  else if (nil (next)) return val;
  else if (next->read (var, val)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    me->bindings (var)= val;
    return val;
  }
  else return val;
}

list<string>
environment_rep::strings_for_completion () const {
  list<string> vars;
  iterator<generic> tmp = entries (bindings);
  for ( ; busy (tmp); ++tmp)
    if (is<literal> (*tmp)) 
      vars = cons (*as<literal>(*tmp), vars);
  return vars;
}

/******************************************************************************
* Converters
******************************************************************************/

inline generic tag (const generic& name, nat id) {
  return gen (name, as<generic> (id)); }
inline generic tag (const generic& name, nat id1, nat id2) {
  return gen (name, as<generic> (id1), as<generic> (id2)); }

void
environment_rep::ensure_up_to_date () const {
  if (!nil (next)) {
    next->ensure_up_to_date ();
    if (next_serial != next->serial) {
      environment_rep* me= const_cast<environment_rep*> (this);
      me->serial++;
      me->next_serial= next->serial;
    }
  }
}

void
environment_rep::set_converter (nat src, nat dest, const generic& val,
				nat trv, nat pen)
{
  //cout << "Set converter ";
  //lisp_print (cout, type_name (src));
  //cout << " -> ";
  //lisp_print (cout, type_name (dest));
  //cout << "\n";

  // add the converter
  set (tag ("converter" , src, dest), val);
  set (tag ("transitive", src, dest), as<generic> (trv));
  set (tag ("penalty", src, dest), as<generic> (pen));

  // add the converter to the list of converters
  generic t= tag ("converters", src);
  if (!contains (t)) set (t, as<generic> (array<nat> ()));
  set (t, as<generic> (cons<nat> (dest, as<array<nat> > (get (t)))));

  // time stamp
  environment_rep* me= const_cast<environment_rep*> (this);
  me->serial++;
}

generic
environment_rep::get_converter (nat src, nat dest, nat& pen) const {
  //cout << "Convert: " << src << ", " << dest << "\n";
  if (src == dest) {
    pen= 0;
    return as<generic> (identity_routine (array<nat> (dest, src)));
  }
  if (dest == 0) {
    if (is_alias_type (src)) {
      routine r;
      alias_reader (src, r);
      pen= PENALTY_AUTOMATIC;
      return as<generic> (r);
    }
    else {
      pen= PENALTY_AUTOMATIC;
      return as<generic> (identity_routine (array<nat> (dest, src)));
    }
  }

  ensure_up_to_date ();
  generic stamp;
  bool ok= read (tag ("cache/converters", src), stamp);
  if (!ok || !is<nat> (stamp) || serial != as<nat> (stamp)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    table<routine,nat> funt;
    table<nat,nat> trvt;
    table<nat,nat> pent;
    array<nat> todo;

    // setup identity converter
    routine id= identity_routine (array<nat> (src, src));
    me->set (tag ("cache/converter", src, src), as<generic> (id));
    funt (src)= id;
    trvt (src)= 3;
    pent (src)= 0;
    todo << src;

    // setup converter to generic
    if (src != 0) {
      if (is_alias_type (src)) alias_reader (src, id);
      else id= identity_routine (array<nat> ((nat) 0, src));
      me->set (tag ("cache/converter", (nat) 0, src), as<generic> (id));
      funt (0)= id;
      trvt (0)= 3;
      pent (0)= PENALTY_PROMOTE_GENERIC;
      todo << 0;
    }

    // transitive closure
    for (nat i=0; i<N(todo); i++) {
      nat cur= todo[i];
      generic cvs;
      if (read (tag ("converters", cur), cvs)) {
	generic all= gen ("converters", as<generic> (cur));
	array<nat> succ= as<array<nat> > (cvs);
	for (nat j=0; j<N(succ); j++) {
	  nat     next    = succ[j];
	  generic next_fun= get (tag ("converter", cur, next));
	  generic next_trv= get (tag ("transitive", cur, next));
	  generic next_pen= get (tag ("penalty", cur, next));
	  assert (is<routine> (next_fun));
	  assert (is<nat> (next_trv));
	  assert (is<nat> (next_pen));
	  nat trv1= trvt[cur], trv2= as<nat> (next_trv);
	  nat pen1= pent[cur], pen2= as<nat> (next_pen);
	  nat trv= (trv1&1) + (trv2&2);
	  nat pen= max (pen1, pen2);
	  if (((trv1&2) != 0 || (trv2&1) != 0) &&
	      (!funt->contains (next) || pen <= pent[next]) &&
	      (pen < pent[next] || (trv & (~trvt[next])) > 0))
	    {
	      routine fun= compose (as<routine> (next_fun), funt[cur]);
	      me->set (tag ("cache/converter", src, next), as<generic> (fun));
	      me->set (tag ("cache/penalty", src, next), as<generic> (pen));
	      funt (next)= fun;
	      trvt (next)= trv;
	      pent (next)= pen;
	      todo << next;
	      //cout << "Converter " << type_name (src) << " -> "
	      //<< type_name (next) << ": " << pen << "\n";
	    }
	}
      }
    }

    // time stamp
    me->set (tag ("cache/converters", src), as<generic> (serial));
  }

  generic fun;
  if (read (tag ("cache/converter", src, dest), fun)) {
    generic penalty;
    read (tag ("cache/penalty", src, dest), penalty);
    pen= as<nat> (penalty);
    return fun;
  }
  pen= PENALTY_INVALID;
  return as<generic> (routine ());
}

} // namespace mmx
