
/******************************************************************************
* MODULE     : language_glue.cpp
* DESCRIPTION: Mathemagix language constructs
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "category_glue.hpp"
#include "base_evaluator.hpp"
#include "generic_internal.hpp"
#include "compound.hpp"
#include "literal.hpp"
#include "exception.hpp"
#include "tuple.hpp"
#include "basix_utils.hpp"
namespace mmx {

/******************************************************************************
* Void value
******************************************************************************/

inline generic
void_value () {
  static generic r= as<generic> (tuple<generic> (gen (MMX_TUPLE)));
  return r;
}

/******************************************************************************
* Transtyping
******************************************************************************/

generic
convert_to (const generic& r, nat t_id, const generic& where) {
  nat r_id= type (r);
  if (r_id == type_id<alias<generic> > ())
    return convert_to (specialize_alias (r), t_id, where);
  if (r_id == t_id) return r;
  else {
    nat pen;
    generic cv=
      current_ev->get_environment ()->get_converter (r_id, t_id, pen);
    if (is<routine> (cv) && pen < PENALTY_INVALID)
      return as<routine> (cv) -> apply (r);
    else return type_mismatch (type_name (t_id), where);
  }
}

generic
mmx_transtype (const generic& g) {
  if (N(g) != 3) return wrong_nr_args (g);
  generic r= eval (g[1]);
  if (is<exception> (r)) return r;
  generic t= eval (g[2]);
  if (is<exception> (t)) return t;
  nat t_id= type_id (t);
  if (t_id == 1) return type_mismatch ("Type", g[2]);
  return convert_to (r, t_id, g[1]);
}

template<typename C> generic
eval_as (const generic& g) {
  generic r= eval (g);
  if (is<C> (r) || is<exception> (r)) return r;
  nat t_id= type_id<C> ();
  return convert_to (r, t_id, g);
}

generic
mmx_type (const generic& g) {
  if (N(g) != 2) return wrong_nr_args (g);
  generic r= eval (g[1]);
  if (is<exception> (r)) return r;
  return type_name (type (r));
}

/******************************************************************************
* Begin statement
******************************************************************************/

generic
mmx_begin (const generic& x) {
  nat i, n= N(x);
  if (n==1) return void_value ();
  for (i=1; i<n-1; i++) {
    generic aux= eval (x[i]);
    if (is<exception> (aux)) return aux;
  }
  return eval (x[i]);
}

/******************************************************************************
* Quatified definitions
******************************************************************************/

generic
mmx_forall (const generic& x) {
  if (N(x) != 3) return wrong_nr_args (x);
  if (N(x[1]) == 0) return eval (x[2]);
  if (N(x[1]) > 1)
    return eval (gen (x[0], gen (x[1][0]), gen (x[0], cdr (x[1]), x[2])));
  generic var= x[1][0];
  if (is_func (var, MMX_TYPE, 2)) var= var[1];
  if (!is<literal> (var)) return type_mismatch ("Literal", var);
  nat ok= 0;
  generic last;
  array<generic> tv= all_type_names ();
  for (nat i=0; i<N(tv); i++) {
    if (current_ev->contains (var)) {
      generic old= current_ev->get (var);
      current_ev->set (var, tv[i]);
      last= eval (x[2]);
      current_ev->set (var, old);
    }
    else {
      current_ev->set (var, tv[i]);
      last= eval (x[2]);
      current_ev->reset (var);
    }
    if (is<exception> (last)) {
      generic msg= *as<exception> (last);
      if (!is_func (msg, "error", 3) ||
	  neq (msg[1], "expected") ||
	  neq (msg[2], "Type"))
	return msg;
    }
    else ok++;
  }
  if (ok == 0) return last;
  return void_value ();
}

/******************************************************************************
* Assignment
******************************************************************************/

static generic
perform_op (nat op, const generic& x, const generic& y) {
  switch (op) {
  case 1: return x+y;
  case 2: return x-y;
  case 3: return x*y;
  case 4: return x/y;
  default: assert (false);
  }
}

static generic
mmx_op_assign (const generic& x, nat op) {
  generic var= eval (x[1]);
  if (is<exception> (var)) return var;
  if (is_alias_type (type (var))) {
    generic val= eval (x[2]);
    if (is<exception> (val)) return val;
    if (op != 0) {
      val= perform_op (op, read_alias (var), val);
      if (is<exception> (val)) return val;
    }
    val= convert_to (val, alias_to_scalar (type (var)), x[2]);
    if (is<exception> (val)) return val;
    return write_alias (var, val);
  }
  if (is_tuple_type (type (var))) {
    generic val= eval (x[2]);
    if (is<exception> (val)) return val;
    if (!is_tuple_type (type (val)))
      return type_mismatch ("Tuple", x[1]);
    generic vars= *as<tuple<generic> > (var);
    generic vals= *as<tuple<generic> > (val);
    if (N(vars) != N(vals))
      return wrong_nr_args (x[2]);
    array<generic> ret= array_fill<generic> (MMX_TUPLE, N(vars));
    for (nat i=1; i<N(vars); i++) {
      nat tid= type (vars[i]);
      if (!is_alias_type (tid))
	return type_mismatch ("Tuple (Alias)", x[1]);
      generic val= vals[i];
      if (op != 0) {
	val= perform_op (op, read_alias (vars[i]), val);
	if (is<exception> (val)) return val;
      }
      val= convert_to (val, alias_to_scalar (tid), x[2]);
      if (is<exception> (val)) return val;
      ret(i)= val;
    }
    for (nat i=1; i<N(vars); i++)
      ret[i]= write_alias (vars[i], ret[i]);
    return as<generic> (tuple<generic> (array_to_compound (ret)));
  }
  return type_mismatch ("Alias or Tuple", x[1]);
}

generic
mmx_plus_assign (const generic& x) {
  if (N(x) != 3) return wrong_nr_args (x);
  return mmx_op_assign (x, 1);
}

generic
mmx_minus_assign (const generic& x) {
  if (N(x) != 3) return wrong_nr_args (x);
  return mmx_op_assign (x, 2);
}

generic
mmx_times_assign (const generic& x) {
  if (N(x) != 3) return wrong_nr_args (x);
  return mmx_op_assign (x, 3);
}

generic
mmx_over_assign (const generic& x) {
  if (N(x) != 3) return wrong_nr_args (x);
  return mmx_op_assign (x, 4);
}

/******************************************************************************
* Definitions
******************************************************************************/

generic
mmx_define (const generic& x) {
  if (N(x) != 3) return wrong_nr_args (x);
  if (is_func (x[1], MMX_TYPE, 2)) {
    if (is<compound> (x[1][1])) {
      generic lambda= gen ("lambda", x[1][1], x[2], x[1][2]);
      return mmx_define (gen (MMX_DEFINE, car (x[1][1]), lambda));
    }
    else {
      if (!is<literal> (x[1][1]))
	return type_mismatch ("Symbol", x[1][1]);
      generic r= eval (gen (MMX_TRANSTYPE, x[2], x[1][2]));
      if (!is<exception> (r))
	current_ev->overload (x[1][1], r);
      return r;
    }
  }
  else {
    if (is<compound> (x[1])) {
      generic lambda= gen ("lambda", x[1], x[2]);
      return mmx_define (gen (MMX_DEFINE, car (x[1]), lambda));
    }
    else {
      if (!is<literal> (x[1][1]))
	return type_mismatch ("Literal", x[1][1]);
      generic r= eval (x[2]);
      if (!is<exception> (r))
	current_ev->overload (x[1], r);
      return r;
    }
  }
}

generic
mmx_assign (const generic& x) {
  if (N(x) != 3) return wrong_nr_args (x);
  if (is_func (x[1], MMX_TYPE, 2)) {
    if (!is<literal> (x[1][1]))
      return type_mismatch ("Symbol", x[1][1]);
    generic r= eval (x[2]);
    if (is<exception> (r)) return r;
    generic t= eval (x[1][2]);
    if (is<exception> (t)) return t;
    nat tid= type_id (t);
    if (tid == 1) return type_mismatch ("Type", x[1][2]);
    if (tid == 0)
      current_ev->set (x[1][1], as<generic> (alias<generic> (r)));
    else {
      r= convert_to (r, tid, x[2]);
      if (is<exception> (r)) return r;
      current_ev->set (x[1][1], current_ev->apply ("alias", r));
    }
    return r;
  }
  else if (is<literal> (x[1]) && !current_ev->contains (x[1])) {
    generic var= gen (MMX_TYPE, x[1], "Generic");
    return mmx_assign (gen (MMX_ASSIGN, var, x[2]));
  }
  else return mmx_op_assign (x, 0);
}

/******************************************************************************
* Lambda expressions
******************************************************************************/

class closure_routine_rep: public routine_rep {
  evaluator  ev;
  array<nat> sig;
  generic    args;
  generic    body;
public:
  closure_routine_rep (const evaluator& e, const array<nat>& s,
		       const generic& a, const generic& b):
    routine_rep ("closure"), ev (e), sig (s), args (a), body (b) {}
  generic xapply (const generic& x) const {
    nat i, n= N(args);
    if (N(x) != n) return wrong_nr_args (x);
    select_evaluator (base_evaluator (ev));
    for (i=1; i<n; i++) {
      if (sig[i] == 0) current_ev->set (args[i], x[i]);
      else {
	generic r= convert_to (x[i], sig[i], args[i]);
	if (is<exception> (r)) {
	  restore_evaluator ();
	  return r;
	}
	current_ev->set (args[i], r);
      }
    }
    generic r= eval (body);
    if (is<exception> (r)) {
      generic msg= *as<exception> (r);
      if (is_func (msg, "return", 1)) r= msg[1];
    }
    if (sig[0] != 0 && !is<exception> (r))
      r= convert_to (r, sig[0], body);
    restore_evaluator ();
    return r;
  }
  array<nat> signature () const { return sig; }
};

routine
closure (const evaluator& ev, const array<nat>& sig,
	 const generic& args, const generic& body)
{
  return new closure_routine_rep (ev, sig, args, body);
}

generic
mmx_lambda (const generic& x) {
  if (N(x) < 3 || N(x) > 4) return wrong_nr_args (x);
  array<generic> args= compound_to_array (x[1]);
  array<nat> sig= array_fill<nat> ((nat) 0, N(args));
  for (nat i=1; i<N(args); i++)
    if (is_func (args[i], MMX_TYPE)) {
      generic t= eval (args[i][2]);
      if (is<exception> (t)) return t;
      nat tid= type_id (t);
      if (tid == 1) return type_mismatch ("Type", args[i][2]);
      sig  (i)= tid;
      args (i)= args[i][1];
    }
  if (N(x) == 4) {
    generic t= eval (x[3]);
    if (is<exception> (t)) return t;
    nat tid= type_id (t);
    if (tid == 1) return type_mismatch ("Type", x[3]);
    sig (0)= tid;
  }
  generic vars= array_to_compound (args);
  return as<generic> (closure (current_ev, sig, vars, x[2]));
}

generic
mmx_mapsto (const generic& x) {
  if (N(x) != 3) return wrong_nr_args (x);
  array<generic> args= array<generic> (generic (MMX_TUPLE));
  if (is_func (x[1], MMX_TUPLE)) args << cdr (compound_to_array (x[1]));
  else args << x[1];
  generic vars= array_to_compound (args);
  if (is_func (x[2], MMX_TYPE, 2))
    return mmx_lambda (gen (MMX_LAMBDA, vars, x[2][1], x[2][2]));
  else return mmx_lambda (gen (MMX_LAMBDA, vars, x[2]));
}

generic
mmx_return (const generic& x) {
  if (N(x) != 2) return wrong_nr_args (x);
  generic ret= eval (x[1]);
  if (is<exception> (ret)) return ret;
  else return as<generic> (exception (gen ("return", ret)));
}

/******************************************************************************
* Boolean operations
******************************************************************************/

generic
mmx_seqand (const generic& x) {
  if (N(x) != 3) return wrong_nr_args (x);
  generic r= eval_as<bool> (x[1]);
  if (is<exception> (r) || !as<bool> (r)) return r;
  return eval_as<bool> (x[2]);
}

generic
mmx_seqor (const generic& x) {
  if (N(x) != 3) return wrong_nr_args (x);
  generic r= eval_as<bool> (x[1]);
  if (is<exception> (r) || as<bool> (r)) return r;
  return eval_as<bool> (x[2]);
}

/******************************************************************************
* If statement
******************************************************************************/

generic
mmx_if (const generic& x) {
  if (N(x) < 3 || N(x) > 4) return wrong_nr_args (x);
  generic cond= eval_as<bool> (x[1]);
  if (is<exception> (cond)) return cond;
  if (as<bool> (cond)) return eval (x[2]);
  else if (N(x) == 4) return eval (x[3]);
  else return void_value ();
}

/******************************************************************************
* Loop statement
******************************************************************************/

generic
mmx_loop (const generic& x) {
  nat i, n= N(x)-1;
  if (n < 1) return wrong_nr_args (x);
  bool stop= false;
  generic aux, ret= void_value ();
  select_evaluator (base_evaluator (current_ev));

  table<iterator<generic>, generic> its;
  table<nat, generic> tps (0);
  for (i=1; i<n; i++)
    if (is_func (x[i], MMX_FOR, 1) && is_func (x[i][1], MMX_IN, 2)) {
      generic var= x[i][1][1];
      if (is_func (var, MMX_TYPE, 2)) {
	generic tp= eval (var[2]);
	if (is<exception> (tp)) { ret= tp; stop= true; break; }
	nat tid= type_id (tp);
	if (tid == 1) {
	  ret= type_mismatch ("Type", var[2]); stop= true; break; }
	var= var[1];
	tps (var)= tid;
      }
      generic val= eval_as<iterator<generic> > (x[i][1][2]);
      if (is<exception> (val)) { ret= val; stop= true; break; }
      its (var)= as<iterator<generic> > (val);
    }
    else if (is_func (x[i], MMX_FOR, 1)) {
      aux= eval (x[i][1]);
      if (is<exception> (aux)) {
	stop= true;
	ret= aux;
	break;
      }
    }

  while (!stop) {
    for (i=1; i<n; i++)
      if (is_func (x[i], MMX_FOR, 1) && is_func (x[i][1], MMX_IN, 2)) {
	generic var= x[i][1][1];
	bool typed= is_func (var, MMX_TYPE, 2);
	if (typed) var= var[1];
	iterator<generic>& it= its (var);
	if (done (it)) { stop= true; break; }
	aux= *it; ++it;
	if (typed) aux= convert_to (aux, tps[var], x[i][1][2]);
	if (is<exception> (aux)) { ret= aux; stop= true; break;	}
	current_ev->set (var, aux);
      }
      else if (is_func (x[i], MMX_WHILE, 1)) {
	generic cond= eval_as<bool> (x[i][1]);
	if (is<exception> (cond)) { ret= cond; stop= true; break; }
	if (!as<bool> (cond)) { stop= true; break; }
      }
    if (stop) break;

    aux= eval (x[n]);
    if (is<exception> (aux)) {
      generic msg= *as<exception> (aux);
      if (eq (msg, gen ("continue")));
      else if (eq (msg, gen ("break"))) break;
      else { ret= aux; break; }
    }

    for (i=1; i<n; i++)
      if (is_func (x[i], MMX_UNTIL, 1)) {
	generic cond= eval_as<bool> (x[i][1]);
	if (is<exception> (cond)) { ret= cond; stop= true; break; }
	if (as<bool> (cond)) { stop= true; break; }
      }
      else if (is_func (x[i], MMX_STEP, 1)) {
	aux= eval (x[i][1]);
	if (is<exception> (aux)) { ret= aux; stop= true; break;	}
      }
  }

  restore_evaluator ();
  return ret;
}

generic
mmx_break (const generic& x) {
  if (N(x) != 1) return wrong_nr_args (x);
  return as<generic> (exception (gen ("break")));
}

generic
mmx_continue (const generic& x) {
  if (N(x) != 1) return wrong_nr_args (x);
  return as<generic> (exception (gen ("continue")));
}

/******************************************************************************
* Range generation
******************************************************************************/

generic
mmx_range (const generic& start, const generic& end) {
  return as<generic> (range_iterator<generic> (start, end, 1, true));
}

generic
mmx_rangeeq (const generic& start, const generic& end) {
  int step= (start<=end? 1: -1);
  return as<generic> (range_iterator<generic> (start, end, step, false));
}

/******************************************************************************
* Where generators
******************************************************************************/

class extract_iterator_rep: public iterator_rep<generic> {
  evaluator ev;
  generic var;
  iterator<generic> it;
  generic cond;
  generic body;
  generic value;

  void spool () {
    select_evaluator (ev);
    while (busy (it)) {
      generic val= *it;
      current_ev->set (var, val);
      generic ok= eval_as<bool> (cond);
      if (is<exception> (ok)) {
	value= ok;
	break;
      }
      else if (as<bool> (ok)) {
	value= eval (body);
	break;
      }
      else ++it;
    }
    restore_evaluator ();
    if (done (it)) value= void_value ();
  }

public:
  extract_iterator_rep (const evaluator& e, const generic& v,
			const iterator<generic>& i, const generic& c,
			const generic& b):
      ev (e), var (v), it (i), cond (c), body (b) { spool (); }
  ~extract_iterator_rep () {}

protected:
  bool is_busy () { return busy (it); }
  void advance () { ++it; spool(); }
  generic current () { return value; }
  iterator_rep<generic>* clone () {
    extract_iterator_rep* rep=
      new extract_iterator_rep (ev, var, it, cond, body);
    rep->value= value;
    return rep;
  }
};


inline iterator<generic>
extract_iterator (const evaluator& ev, const generic& var,
		  const iterator<generic>& it, const generic& cond,
		  const generic& v)
{
  return iterator<generic> (new extract_iterator_rep (ev, var, it, cond, v));
}

class unnest_iterator_rep: public iterator_rep<generic> {
  iterator<generic> it;
  iterator<generic> subit;

  void spool () {
    while (busy (it)) {
      generic val= *it; ++it;
      if (!is<exception> (val) && !is<iterator<generic> > (val))
	val= type_mismatch (gen ("Generator", generic ("Generic")), val);
      if (is<exception> (val)) subit= seq<generic> (val);
      else subit= as<iterator<generic> > (val);
      if (busy (subit)) break;
    }
  }

public:
  unnest_iterator_rep (const iterator<generic>& i): it (i) { spool (); }
  ~unnest_iterator_rep () {}

protected:
  bool is_busy () { return busy (subit); }
  void advance () { ++subit; if (done (subit)) spool(); }
  generic current () { return *subit; }
  iterator_rep<generic>* clone () {
    unnest_iterator_rep* rep= new unnest_iterator_rep (it);
    rep->subit= subit;
    return rep;
  }
};

inline iterator<generic>
unnest_iterator (const iterator<generic>& it) {
  return iterator<generic> (new unnest_iterator_rep (it));
}

generic
mmx_where (const generic& x) {
  nat i, n= N(x);
  if (n < 3) return wrong_nr_args (x);
  if (is_func (x[1], MMX_IN, 2)) {
    // TODO: generalize syntax when x[1] is a tuple
    // with elements of the form 'variable in generator'.
    return mmx_where (append (gen (x[0], x[1][1], x[1]), range (x, 2, N(x))));
  }
  if (!is_func (x[2], MMX_IN, 2))
    return as<generic> (exception ("'variable in generator' expected", x[2]));
  for (i=3; i<n; i++)
    if (is_func (x[i], MMX_IN, 2))
      break;
  if (i == n) {
    generic cond= "true";
    if (i > 3) {
      cond= x[i-1];
      for (nat j=i-2; j>=3; j--)
	cond= gen (MMX_SEQAND, x[j], cond);
    }
    generic var= x[2][1];
    // TODO: allow for typed variables 'var: Type'
    if (!is<literal> (var)) return type_mismatch ("Literal", var);
    generic val= eval_as<iterator<generic> > (x[2][2]);
    if (is<exception> (val)) return val;
    evaluator ev= base_evaluator (current_ev);
    iterator<generic> it= as<iterator<generic> > (val);
    iterator<generic> r = extract_iterator (ev, var, it, cond, x[1]);
    return as<generic> (r);
  }
  else {
    generic inner= append (gen (x[0], x[1]), range (x, i, n));
    generic outer= append (gen (x[0], inner), range (x, 2, i));
    generic it= mmx_where (outer);
    if (is<exception> (it)) return it;
    return as<generic> (unnest_iterator (as<iterator<generic> > (it)));
  }
}

/******************************************************************************
* Tuples
******************************************************************************/

generic equalize (const generic& name, const generic& x);

generic
mmx_tuple (const generic& x) {
  array<generic> v= array<generic> (generic (MMX_TUPLE));
  for (nat i=1; i<N(x); i++) {
    generic val= eval (x[i]);
    if (is<exception> (val)) return val;
    else v << val;
  }
  generic t= equalize (MMX_TUPLE, array_to_compound (v));
  if (N(t) == 2) return t[1];
  return as<generic> (tuple<generic> (t));
}

generic
mmx_explode (const generic& x) {
  array<generic> v= array<generic> (generic (MMX_TUPLE));
  for (nat i=1; i<N(x); i++) {
    generic val= eval (x[i]);
    if (is<exception> (val)) return val;
    else if (is<iterator<generic> > (val)) {
      iterator<generic> it= as<iterator<generic> > (val);
      while (busy (it)) {
	generic next= *it; ++it;
	if (is<exception> (next)) return next;
	else v << next;
      }
    }
    else v << val;    
  }
  generic t= equalize (MMX_TUPLE, array_to_compound (v));
  if (N(t) == 2) return t[1];
  return as<generic> (tuple<generic> (t));
}

generic
mmx_sqtuple (const generic& x) {
  return eval (gen ("list", cons ("explode", cdr (x))));
}

/******************************************************************************
* Printing
******************************************************************************/

class printer_rep REP_STRUCT {
public:
  generic name;
  ostream& out;
public:
  inline printer_rep (const generic& name2, ostream& out2):
    name (name2), out (out2) {}
};

class printer {
INDIRECT_PROTO (printer, printer_rep)
public:
  inline printer (const generic& name, ostream& out):
    rep (new printer_rep (name, out)) {}
  inline friend bool eq (const printer& p1, const printer& p2) {
    return p1->name == p2->name; }
  inline friend bool neq (const printer& p1, const printer& p2) {
    return p1->name != p2->name; }
  inline bool operator == (const printer& p) const {
    return rep->name == p->name; }
  inline bool operator != (const printer& p) const {
    return rep->name != p->name; }
};
INDIRECT_IMPL (printer, printer_rep)

inline nat hash (const printer& p) { return hash (p->name); }
inline generic flatten (const printer& p) { return flatten (p->name); }
PRINT_SUGAR(,printer)

printer
mmx_print (const printer& p, const generic& x) {
  p->out << x;
  return p;
}

/******************************************************************************
* Loading dynamic libraries
******************************************************************************/

generic
mmx_supports (const generic& x) {
  if (N(x) != 2) return wrong_nr_args (x);
  generic r= eval_as<string> (x[1]);
  if (is<exception> (r) || !is<string> (r)) return r;
  string name= as<string> (r);
  return as<generic> (dl_exists (name));
}

generic
mmx_use (const generic& x) {
  if (N(x) != 2) return wrong_nr_args (x);
  generic r= eval_as<string> (x[1]);
  if (is<exception> (r) || !is<string> (r)) return r;
  string name= as<string> (r);
  string message= dl_link (name);
  if (message != "") return as<generic> (exception (message, x));
  return void_value ();
}

/******************************************************************************
* For testing
******************************************************************************/

generic
mmx_type_name (const int& i) {
  nat id= (nat) i;
  return type_name (id);
}

/******************************************************************************
* Interface
******************************************************************************/

string
make_literal_string (const literal& lit) {
  string s= *lit;
  return unquote (s);
}

void
define_language () {
  static printer mmout ("mmout", cout);
  static printer mmerr ("mmerr", cerr);
  define_type<always,string> ("String");
  define_type<always,printer> ("Printer");
  define_type<always,exception> ("Exception");
  define_type<always,iterator<generic> > (gen("Generator",generic("Generic")));
  define<always> ("literal_string", make_literal_string);
  define_constant<always,printer> ("mmout", mmout);
  define_constant<always,printer> ("mmerr", mmerr);
  define<always> (MMX_RANGE, mmx_range);
  define<always> (MMX_RANGEEQ, mmx_rangeeq);
  define<always> (MMX_LESSLESS, mmx_print);
  define<always> (MMX_MELT, (generic (*) (const generic&)) melt);
  define<always> (MMX_FUSE, (generic (*) (const generic&)) fuse);
  define_primitive<always> (MMX_TRANSTYPE, mmx_transtype);
  define_primitive<always> ("type", mmx_type);
  define_primitive<always> (MMX_BEGIN, mmx_begin);
  define_primitive<always> (MMX_FORALL, mmx_forall);
  define_primitive<always> (MMX_DEFINE, mmx_define);
  define_primitive<always> (MMX_ASSIGN, mmx_assign);
  define_primitive<always> (MMX_PLUS_ASSIGN, mmx_plus_assign);
  define_primitive<always> (MMX_MINUS_ASSIGN, mmx_minus_assign);
  define_primitive<always> (MMX_TIMES_ASSIGN, mmx_times_assign);
  define_primitive<always> (MMX_OVER_ASSIGN, mmx_over_assign);
  define_primitive<always> (MMX_LAMBDA, mmx_lambda);
  define_primitive<always> (MMX_MAPSTO, mmx_mapsto);
  define_primitive<always> (MMX_RETURN, mmx_return);
  define_primitive<always> (MMX_SEQAND, mmx_seqand);
  define_primitive<always> (MMX_SEQOR, mmx_seqor);
  define_primitive<always> (MMX_IF, mmx_if);
  define_primitive<always> (MMX_LOOP, mmx_loop);
  define_primitive<always> (MMX_BREAK, mmx_break);
  define_primitive<always> (MMX_CONTINUE, mmx_continue);
  define_primitive<always> (MMX_WHERE, mmx_where);
  define_primitive<always> (MMX_VWHERE, mmx_where);
  define_primitive<always> (MMX_TUPLE, mmx_tuple);
  define_primitive<always> (MMX_SQTUPLE, mmx_sqtuple);
  define_primitive<always> ("explode", mmx_explode);
  define_primitive<always> ("supports?", mmx_supports);
  define_primitive<always> ("use", mmx_use);
  define<always> ("type_name", mmx_type_name);
}

} // namespace mmx
